#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/LabelG.h>
#include <Xm/PushBG.h>
#include <Xm/List.h>
#include <Xm/ToggleBG.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include "global.h"
#include "messages.h"

#define  NTRANSFORM1 7
#define  NTRANSFORM2 6
#define  NTRANSMAX   7  /* set to maximum of NTRANSFORM1 and NTRANSFORM2 */

/* Structure for one-variable transformations */
typedef struct {
  int  *variables;
  int  transform;
  char *paraa, *parab;
} trans1_struct;

typedef struct {
  int   transform;
  int   *list1, *list2;
  char  **rlist;
  int   nlist1, nlist2, nrlist;
} trans2_struct;

void trans_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Widget        toplevel, dialog, form, frame1, frame2, var_w, radio_box, toggle[NTRANSMAX];
  Widget        label_a, label_b, para_a, para_b;
  Widget        res_w, eq_lab, var1_w, var2_w, frame3, frame4, frame5;
  Widget        del_but1, del_but2, del_but3, add_but1, add_but2;
  Widget        t_field, all_w;
  Arg           args[9];
  XmStringTable var_list;
  String        *lcol;
  int           item, i;
  int           cols;
  static trans1_struct *trans;
  static trans2_struct *trans2;

  extern void showmess(Widget widget, char *title, char *text);
  extern void cancel_cb(Widget, XtPointer, XtPointer);
  void trans1_ok_cb(Widget, XtPointer, XtPointer);
  void var_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void toggle_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void para_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void addtolist_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void addtores_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void delfromlist_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void toggle2_cb(Widget widget, XtPointer client_data, XtPointer call_data);
  void trans2_ok_cb(Widget, XtPointer, XtPointer);

  /* allocate trans1_struct */
  if (!trans) {
    trans            = new trans1_struct;
    trans->transform = 0;
    trans->variables = new int[2];
    trans->paraa     = new char[10];
    strcpy(trans->paraa, "1");
    trans->parab     = new char[10];
    strcpy(trans->parab, "0");
  }

  /* allocate trans2_struct */
  if (!trans2) {
    trans2 = new trans2_struct;
    trans2->transform = 0;
    trans2->nlist1    = 0;
    trans2->list1     = new int[1];
    trans2->nlist2    = 0;
    trans2->list2     = new int[1];
    trans2->nrlist    = 0;
    trans2->rlist     = new char*[1];
  }

  /* get various data structurs */
  toplevel=XtParent(widget);
  while (!XtIsTopLevelShell(toplevel)) toplevel=XtParent(toplevel);
  XtVaGetValues(StatData, XmNcolumns, &cols, NULL);
  XtVaGetValues(StatData, XmNcolumnLabels, &lcol, NULL);
  
  if (cols>0) trans->variables=(int *)realloc(trans->variables, cols*sizeof(int));
  for (i=0; i<cols; i++) trans->variables[i]=0;

  trans2->nlist1=0;
  trans2->nlist2=0;
  trans2->nrlist=0;
  trans2->transform=0;

  item=(int) client_data;
  switch(item) {
  case 0:      // one-variable transformations;
    i=0;
    XtSetArg(args[i], XmNdialogTitle,
	     XmStringCreateLocalized("one-variable transformations")); i++;
    XtSetArg(args[i], XmNmessageString,
	     XmStringCreateLocalized("select variables, transformation, and parameters:")); i++;
    XtSetArg(args[i], XmNautoUnmanage, False); i++;
    XtSetArg(args[i], XmNuserData, trans); i++;
    dialog=XmCreateMessageDialog(toplevel, "transdialog", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    XtVaSetValues(dialog, XmNheight, 400, NULL);
    form=XtVaCreateWidget("form", xmFormWidgetClass, dialog, NULL);
    /* create textfields for parameters a and b */
    label_b=XtVaCreateManagedWidget("parameter b:", xmLabelGadgetClass, form,
				    XmNbottomAttachment, XmATTACH_FORM,
				    XmNleftAttachment, XmATTACH_FORM,
				    NULL);
    para_b=XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
				   XmNbottomAttachment, XmATTACH_FORM,
				   XmNrightAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, label_b,
				   XmNuserData, trans,
				   NULL);
    XmTextSetString(para_b, trans->parab);
    XtAddCallback(para_b, XmNlosingFocusCallback, para_cb, (XtPointer) 1);
    label_a=XtVaCreateManagedWidget("parameter a:", xmLabelGadgetClass, form,
				    XmNbottomAttachment, XmATTACH_WIDGET,
				    XmNbottomWidget, para_b,
				    XmNleftAttachment, XmATTACH_FORM,
				    NULL);
    para_a=XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget, para_b,
				   XmNrightAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, label_a,
				   XmNuserData, trans,
				   NULL);
    XmTextSetString(para_a, trans->paraa);
    XtAddCallback(para_a, XmNlosingFocusCallback, para_cb, (XtPointer) 0);
    /* create radiobox for transformations */
    frame2=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget, para_a,
				   XmNrightAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, label_a,
				   XmNshadowType, XmSHADOW_OUT,
				   NULL);
    XtVaCreateManagedWidget("transformations", xmLabelGadgetClass, frame2,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    radio_box=XmCreateRadioBox(frame2, "radio_box", NULL, 0);
    XtVaSetValues(radio_box,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);
    toggle[0]=XtVaCreateManagedWidget("a * x + b",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[1]=XtVaCreateManagedWidget("lg(a * x + b)",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[2]=XtVaCreateManagedWidget("ln(a * x + b)",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[3]=XtVaCreateManagedWidget("exp(a * x + b)",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[4]=XtVaCreateManagedWidget("(a * x + b) ^ 2",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[5]=XtVaCreateManagedWidget("SQRT(a * x + b)",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    toggle[6]=XtVaCreateManagedWidget("1/(a * x + b)",
				      xmToggleButtonGadgetClass, radio_box,
				      XmNuserData, trans,
				      NULL);
    XtVaSetValues(toggle[trans->transform], XmNset, True, NULL);
    for (i=0; i<NTRANSFORM1; i++) 
      XtAddCallback(toggle[i], XmNvalueChangedCallback, toggle_cb, (XtPointer) i);
    XtManageChild(radio_box);
    /* create var-list */
    if (cols>0) {
      var_list=(XmStringTable) XtMalloc (cols * sizeof(XmString));
      for (i=0; i<cols; i++) {
	var_list[i]=XmStringCreateLocalized(lcol[i]);
      }
    }
    frame1=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNwidth, 150,
				   XmNshadowType, XmSHADOW_OUT,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget, para_a,
				   XmNrightAttachment, XmATTACH_WIDGET,
				   XmNrightWidget, frame2,
				   XmNleftAttachment, XmATTACH_FORM,
				   NULL);
    XtVaCreateManagedWidget("variables", xmLabelGadgetClass, frame1,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    var_w=XmCreateScrolledList(frame1, "varlist", NULL, 0);
    XtVaSetValues(var_w,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  XmNuserData, trans,
		  NULL);
    if (cols>0) {
      XtVaSetValues(var_w,
		    XmNitemCount, cols,
		    XmNitems, var_list,
		    NULL);
    }
    XtManageChild(var_w);
    XtAddCallback(var_w, XmNextendedSelectionCallback, var_cb, NULL);
    XtManageChild(form);
    XtManageChild(dialog);
    XtAddCallback(dialog, XmNokCallback, trans1_ok_cb, NULL);
    XtAddCallback(dialog, XmNcancelCallback, cancel_cb, NULL);
    XtPopup(XtParent(dialog), XtGrabNone);

    if (cols > 0) {
      for (i=0; i<cols; i++) XmStringFree(var_list[i]);
      XtFree((char *)var_list);
    }
    break;
  case 1: /* two-variables transformations */
    i=0;
    XtSetArg(args[i], XmNdialogTitle,
	     XmStringCreateLocalized("two-variable transformations")); i++;
    XtSetArg(args[i], XmNautoUnmanage, False); i++;
    XtSetArg(args[i], XmNokLabelString, XmStringCreateLocalized("OK")); i++;
    XtSetArg(args[i], XmNcancelLabelString, XmStringCreateLocalized("Cancel")); i++;
    XtSetArg(args[i], XmNuserData, trans2); i++;
    XtSetArg(args[i], XmNdefaultButtonType, XmDIALOG_NONE); i++;
    dialog=XmCreateTemplateDialog(toplevel, "transdialog", args, i);
    form=XtVaCreateWidget("form", xmFormWidgetClass, dialog, NULL);
    /* create various widgets */
    /* result variables list  */
    frame1=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNshadowType, XmSHADOW_OUT,
				   XmNwidth, 150,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_FORM,
				   NULL);
    XtVaCreateManagedWidget("result variable", xmLabelGadgetClass, frame1,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    res_w=XmCreateScrolledList(frame1, "varlist", NULL, 0);
    XtVaSetValues(res_w,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  XmNvisibleItemCount, 11,
		  XmNuserData, 0,
		  NULL);
    XtManageChild(res_w);
    /* = label */
    eq_lab=XtVaCreateManagedWidget(" = ", xmLabelGadgetClass, form,
			      XmNtopAttachment, XmATTACH_FORM,
			      XmNleftAttachment, XmATTACH_WIDGET,
			      XmNleftWidget, frame1,
			      NULL);
    /* variable 1 list */
    frame2=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNshadowType, XmSHADOW_OUT,
				   XmNwidth, 150,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, eq_lab,
				   NULL);
    XtVaCreateManagedWidget("1st variable     ", xmLabelGadgetClass, frame2,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    var1_w=XmCreateScrolledList(frame2, "varlist", NULL, 0);
    XtVaSetValues(var1_w,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  XmNvisibleItemCount, 11,
		  XmNuserData, 1,
		  NULL);
    XtManageChild(var1_w);
    /* first del button */
    del_but1=XtVaCreateManagedWidget("delete", xmPushButtonGadgetClass, form,
				     XmNtopAttachment, XmATTACH_WIDGET,
				     XmNtopWidget, frame1,
				     XmNleftAttachment, XmATTACH_FORM,
				     XmNrightAttachment, XmATTACH_WIDGET,
				     XmNrightWidget, eq_lab,
				     XmNuserData, trans2,
				     NULL);
    XtAddCallback(del_but1, XmNactivateCallback, delfromlist_cb, res_w);
    /* radio button box */
    frame3=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, frame2,
				   XmNbottomAttachment, XmATTACH_WIDGET,
				   XmNbottomWidget, del_but1,
				   XmNshadowType, XmSHADOW_OUT,
				   NULL);
    XtVaCreateManagedWidget("trans", xmLabelGadgetClass, frame3,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    radio_box=XmCreateRadioBox(frame3, "radio_box", NULL, 0);
    XtVaSetValues(radio_box,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);
    toggle[0]=XtVaCreateManagedWidget("   +   ",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    toggle[1]=XtVaCreateManagedWidget("   -   ",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    toggle[2]=XtVaCreateManagedWidget("   *   ",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    toggle[3]=XtVaCreateManagedWidget("   /   ",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    toggle[4]=XtVaCreateManagedWidget("100*v2/v1",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    toggle[5]=XtVaCreateManagedWidget("100*(v2-v1)/v1",
				      xmToggleButtonGadgetClass, radio_box,
				      NULL);
    XtVaSetValues(toggle[0], XmNset, True, NULL);
    for (i=0; i<NTRANSFORM2; i++) {
      XtAddCallback(toggle[i], XmNvalueChangedCallback, toggle2_cb, (XtPointer) i);
      XtVaSetValues(toggle[i], XmNuserData, trans2, NULL);
    }
    XtManageChild(radio_box);
    /* variable 2 list */
    frame4=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNshadowType, XmSHADOW_OUT,
				   XmNwidth, 150,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, frame3,
				   XmNrightAttachment, XmATTACH_FORM,
				   NULL);
    XtVaCreateManagedWidget("2nd variable     ", xmLabelGadgetClass, frame4,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    var2_w=XmCreateScrolledList(frame4, "varlist", NULL, 0);
    XtVaSetValues(var2_w,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  XmNvisibleItemCount, 11,
		  XmNuserData, 2,
		  NULL);
    XtManageChild(var2_w);
    /* buttons */
    del_but2=XtVaCreateManagedWidget("delete", xmPushButtonGadgetClass, form,
				     XmNtopAttachment, XmATTACH_WIDGET,
				     XmNtopWidget, frame1,
				     XmNleftAttachment, XmATTACH_WIDGET,
				     XmNleftWidget, eq_lab,
				     XmNrightAttachment, XmATTACH_WIDGET,
				     XmNrightWidget, frame3,
				     XmNuserData, trans2,
				     NULL);
    XtAddCallback(del_but2, XmNactivateCallback, delfromlist_cb, var1_w);
    del_but3=XtVaCreateManagedWidget("delete", xmPushButtonGadgetClass, form,
				     XmNtopAttachment, XmATTACH_WIDGET,
				     XmNtopWidget, frame1,
				     XmNleftAttachment, XmATTACH_WIDGET,
				     XmNleftWidget, frame3,
				     XmNrightAttachment, XmATTACH_FORM,
				     XmNuserData, trans2,
				     NULL);
    XtAddCallback(del_but3, XmNactivateCallback, delfromlist_cb, var2_w);
    add_but1=XtVaCreateManagedWidget("add", xmPushButtonGadgetClass, form,
				     XmNtopAttachment, XmATTACH_WIDGET,
				     XmNtopWidget, del_but2,
				     XmNleftAttachment, XmATTACH_WIDGET,
				     XmNleftWidget, eq_lab,
				     XmNrightAttachment, XmATTACH_WIDGET,
				     XmNrightWidget, frame3,
				     NULL);
    XtAddCallback(add_but1, XmNactivateCallback, addtolist_cb, var1_w);
    add_but2=XtVaCreateManagedWidget("add", xmPushButtonGadgetClass, form,
				     XmNtopAttachment, XmATTACH_WIDGET,
				     XmNtopWidget, del_but2,
				     XmNleftAttachment, XmATTACH_WIDGET,
				     XmNleftWidget, frame3,
				     XmNrightAttachment, XmATTACH_FORM,
				     NULL);
    XtAddCallback(add_but2, XmNactivateCallback, addtolist_cb, var2_w);
    /* text field */
    t_field=XtVaCreateManagedWidget("text", xmTextWidgetClass, form,
				    XmNtopAttachment, XmATTACH_WIDGET,
				    XmNtopWidget, del_but2,
				    XmNleftAttachment, XmATTACH_FORM,
				    XmNrightAttachment, XmATTACH_WIDGET,
				    XmNrightWidget, eq_lab,
				    XmNuserData, trans2,
				    NULL);
    XtAddCallback(t_field, XmNactivateCallback, addtores_cb, res_w);
    /* list of all variables */
    if (cols>0) {
      var_list=(XmStringTable) XtMalloc (cols * sizeof(XmString));
      for (i=0; i<cols; i++) {
	var_list[i]=XmStringCreateLocalized(lcol[i]);
      }
    }
    frame5=XtVaCreateManagedWidget("frame", xmFrameWidgetClass, form,
				   XmNshadowType, XmSHADOW_OUT,
				   XmNtopAttachment, XmATTACH_WIDGET,
				   XmNtopWidget, add_but1,
				   XmNleftAttachment, XmATTACH_WIDGET,
				   XmNleftWidget, eq_lab,
				   XmNrightAttachment, XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_FORM,
				   NULL);
    XtVaCreateManagedWidget("select variable", xmLabelGadgetClass, frame5,
			    XmNchildType, XmFRAME_TITLE_CHILD,
			    XmNchildHorizontalAlignment, XmALIGNMENT_BEGINNING,
			    XmNchildVerticalAlignment, XmALIGNMENT_WIDGET_TOP,
			    NULL);
    all_w=XmCreateScrolledList(frame5, "varlist", NULL, 0);
    XtVaSetValues(all_w,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  XmNuserData, trans2,
		  NULL);
    if (cols>0) {
      XtVaSetValues(all_w,
		    XmNitemCount, cols,
		    XmNitems, var_list,
		    NULL);
      if (cols<=10) XtVaSetValues(all_w, XmNvisibleItemCount, cols, NULL);
      else XtVaSetValues(all_w, XmNvisibleItemCount, 10, NULL);
    }
    XtManageChild(all_w);

    /* manage everything */
    XtVaSetValues(add_but1, XmNuserData, all_w, NULL);
    XtVaSetValues(add_but2, XmNuserData, all_w, NULL);
    XtManageChild(form);
    XtManageChild(dialog);
    XtAddCallback(dialog, XmNcancelCallback, cancel_cb, NULL);
    XtAddCallback(dialog, XmNokCallback, trans2_ok_cb, NULL);
    XtPopup(XtParent(dialog), XtGrabNone);

    /* free memory */
    if (cols > 0) {
      for (i=0; i<cols; i++) XmStringFree(var_list[i]);
      XtFree((char *)var_list);
    }
    break;
  }
}

/* ..................................*/
/* .....   C a l l b a c k s   ..... */
/* ................................. */

void var_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  XmListCallbackStruct *cbs=(XmListCallbackStruct *) call_data;
  trans1_struct        *trans;
  int                  i, nitems;
  
  XtVaGetValues(widget, XmNuserData, &trans, NULL);
  XtVaGetValues(widget, XmNitemCount, &nitems, NULL);

  for (i=0; i<nitems; i++) trans->variables[i]=0;
  for (i=0; i<cbs->selected_item_count; i++) 
    trans->variables[cbs->selected_item_positions[i]-1]=1;
}

void toggle_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  trans1_struct  *trans;
  int            which;

  which=(int) client_data;
  XtVaGetValues(widget, XmNuserData, &trans, NULL);

  if (XmToggleButtonGadgetGetState(widget)) 
    trans->transform=which;
}

void para_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int           which;
  trans1_struct *trans;
  char          *text;

  which=(int)client_data;
  XtVaGetValues(widget, XmNuserData, &trans, NULL);
  text=XmTextGetString(widget);

  if (which==0) {  // parameter a;
    trans->paraa=(char *)realloc(trans->paraa, strlen(text)+1);
    strcpy(trans->paraa, text);
  }
  if (which==1) {  // parameter b;
    trans->parab=(char *)realloc(trans->parab, strlen(text)+1);
    strcpy(trans->parab, text);
  }

  XtFree(text);
}


void trans1_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  String        **cells;
  trans1_struct *trans;
  int           i, j, nvars, cols, rows;
  short         is_num, is_valid;
  float         a, b, wert;
  char          *pchar, text[256];

  extern void showmess(Widget widget, char *title, char *text);

  XtVaGetValues(widget, XmNuserData, &trans, NULL);
  XtVaGetValues(StatData, XmNcolumns, &cols, NULL);
  XtVaGetValues(StatData, XmNrows, &rows, NULL);
  XtVaGetValues(StatData, XmNcells, &cells, NULL);

  nvars=0;
  for (i=0; i<cols; i++)
    if (trans->variables[i]) nvars++;

  is_num=1;
  a=strtod(trans->paraa, &pchar);
  if (a==0 && !strcmp(pchar, trans->paraa)) is_num=0;
  b=strtod(trans->parab, &pchar);
  if (b==0 && !strcmp(pchar, trans->parab)) is_num=0;
  
  if (is_num && (nvars>0)) {
    switch(trans->transform) {
    case 0:      // a * x + b;
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      wert=a * wert + b;
	      sprintf(text, "%f", wert);
	      XbaeMatrixSetCell(StatData, j, i, text);
	    }
	  }
      break;
    case 1:      // lg(a * x + b);
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      if ((a * wert+b)!=0) {
		wert=log10(a*wert + b);
		sprintf(text, "%f", wert);
		XbaeMatrixSetCell(StatData, j, i, text);
	      }
	      else XbaeMatrixSetCell(StatData, j, i, "ndef");
	    }
	  }
      break;
    case 2:      // ln(a * x + b);
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      if ((a*wert+b)!=0) {
		wert=log(a*wert + b);
		sprintf(text, "%f", wert);
		XbaeMatrixSetCell(StatData, j, i, text);
	      }
	      else XbaeMatrixSetCell(StatData, j, i, "ndef");
	    }
	  }
      break;
    case 3:      // exp(a * x + b);
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      wert=exp(a * wert + b);
	      sprintf(text, "%f", wert);
	      XbaeMatrixSetCell(StatData, j, i, text);
	    }
	  }
      break;
    case 4:      // (a * x + b) ^ 2;
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      wert=(a * wert + b) * (a * wert + b);
	      sprintf(text, "%f", wert);
	      XbaeMatrixSetCell(StatData, j, i, text);
	    }
	  }
      break;
    case 5:      // SQRT(a * x + b);
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      if (a*wert+b >= 0) {
		wert=sqrt(a * wert + b);
		sprintf(text, "%f", wert);
		XbaeMatrixSetCell(StatData, j, i, text);
	      }
	      else XbaeMatrixSetCell(StatData, j, i, "ndef");
	    }
	  }
      break;
    case 6:      // 1/(a * x + b);
      for (i=0; i<cols; i++) 
	if (trans->variables[i]) 
	  for (j=0; j<rows; j++) {
	    is_valid=1;
	    wert=strtod(cells[j][i], &pchar);
	    if (wert==0 && !strcmp(pchar, cells[j][i])) is_valid=0;
	    if (is_valid) {
	      if (a*wert+b != 0) {
		wert=1/(a * wert + b);
		sprintf(text, "%f", wert);
		XbaeMatrixSetCell(StatData, j, i, text);
	      }
	      else XbaeMatrixSetCell(StatData, j, i, "ndef");
	    }
	  }
      break;
    }
    XtUnmanageChild(widget);
  }
  else {
    if (!is_num && (nvars>0))  showmess(StatData, TRANS1_T, TRANS1_1_M);
    if (!is_num && (nvars<=0)) showmess(StatData, TRANS1_T, TRANS1_2_M);
    if (is_num && (nvars<=0))  showmess(StatData, TRANS1_T, TRANS1_3_M);
  }

}


/* callbacks for two-variable transformations */
void addtolist_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  Widget        var_w, all_w;
  trans2_struct *trans2;
  int           i, which;
  int           *pos_list, pos_cnt;
  XmString      *all_list;

  var_w = (Widget) client_data;
  XtVaGetValues(widget, XmNuserData, &all_w, NULL);
  XtVaGetValues(all_w, XmNuserData, &trans2, NULL);
  XtVaGetValues(var_w, XmNuserData, &which, NULL);
  XmListGetSelectedPos(all_w, &pos_list, &pos_cnt);
  
  if (pos_cnt>0) {
    if (which==1) {
      trans2->nlist1+=pos_cnt;
      trans2->list1=(int *)realloc(trans2->list1, trans2->nlist1*sizeof(int));
      for (i=0; i<pos_cnt; i++) trans2->list1[i+trans2->nlist1-pos_cnt]=pos_list[i]-1;
    }
    if (which==2) {
      trans2->nlist2+=pos_cnt;
      trans2->list2=(int *)realloc(trans2->list2, trans2->nlist2*sizeof(int));
      for (i=0; i<pos_cnt; i++) trans2->list2[i+trans2->nlist2-pos_cnt]=pos_list[i]-1;
    }

    XtVaGetValues(all_w, XmNitems, &all_list, NULL);
    for (i=0; i<pos_cnt; i++) XmListAddItem(var_w, all_list[pos_list[i]-1],0);
    XtFree((char *)pos_list);
  }
}

void addtores_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  char          *varname;
  Widget        res_w;
  trans2_struct *trans2;
  short         is_valid;
  String        *lcol;
  int           i, cols;

  extern void showmess(Widget widget, char *title, char *text);

  res_w=(Widget)client_data;
  XtVaGetValues(widget, XmNuserData, &trans2, NULL);
  XtVaGetValues(StatData, XmNcolumnLabels, &lcol, NULL);
  XtVaGetValues(StatData, XmNcolumns, &cols, NULL);

  if (varname=XmTextGetString(widget)) {
    is_valid=1;
    for (i=0; i<cols; i++) 
      if (!strcmp(varname, lcol[i])) is_valid=0;
    for (i=0; i<trans2->nrlist; i++)
      if (!strcmp(varname, trans2->rlist[i])) is_valid=0;
    if (is_valid) {
      trans2->nrlist++;
      trans2->rlist=(char **)realloc(trans2->rlist, trans2->nrlist*sizeof(char *));
      trans2->rlist[trans2->nrlist-1]= new char[strlen(varname)+1];
      strcpy(trans2->rlist[trans2->nrlist-1], varname);
      XmListAddItem(res_w, XmStringCreateLocalized(varname), 0);
    }
    else showmess(widget, VARNAME_T, VARNAME_M);
    XtFree(varname);
  }
}

void delfromlist_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  int           i, j, which;
  Widget        list_w;
  trans2_struct *trans2;
  int           *pos_list, pos_cnt, *d_list;

  list_w=(Widget)client_data;
  XtVaGetValues(list_w, XmNuserData, &which, NULL);
  XtVaGetValues(widget, XmNuserData, &trans2, NULL);

  XmListGetSelectedPos(list_w, &pos_list, &pos_cnt);
  if (pos_cnt>0) {
    switch(which) {
    case 0:
      d_list=new int[pos_cnt];
      for (i=0; i<pos_cnt; i++) d_list[i]=pos_list[i];
      for (i=0; i<pos_cnt; i++) {
	for (j=d_list[i]-1; j<trans2->nrlist-1; j++) {
	  trans2->rlist[j]=(char *)realloc(trans2->rlist[j], strlen(trans2->rlist[j+1])+1);
	  strcpy(trans2->rlist[j], trans2->rlist[j+1]);
	}
	delete[]trans2->rlist[trans2->nrlist-1];
	trans2->nrlist--;
	if (trans2->nrlist>0)
	  trans2->rlist=(char **)realloc(trans2->rlist, trans2->nrlist*sizeof(char *));
	for (j=i+1; j<pos_cnt; j++) d_list[j]--;
      }
      break;
    case 1:
      for (i=0; i<pos_cnt; i++) trans2->list1[pos_list[i]-1]=-1;
      for (i=0; i<trans2->nlist1; i++) {
	if (trans2->list1[i]==(-1)) {
	  /* delete from trans2->list1 */
	  for (j=i; j<trans2->nlist1-1; j++) {
	    trans2->list1[j]=trans2->list1[j+1];
	  }
	  trans2->nlist1--;
	  if (trans2->nlist1>0)
	    trans2->list1=(int *)realloc(trans2->list1, trans2->nlist1*sizeof(int));
	  i--;
	}
      }
      break;
    case 2:
      for (i=0; i<pos_cnt; i++) trans2->list2[pos_list[i]-1]=-1;
      for (i=0; i<trans2->nlist2; i++) {
	if (trans2->list2[i]==(-1)) {
	  /* delete from trans2->list2 */
	  for (j=i; j<trans2->nlist2-1; j++) {
	    trans2->list2[j]=trans2->list2[j+1];
	  }
	  trans2->nlist2--;
	  if (trans2->nlist2>0)
	    trans2->list2=(int *)realloc(trans2->list2, trans2->nlist2*sizeof(int));
	  i--;
	}
      }
      break;
    }
    XmListDeletePositions(list_w, pos_list, pos_cnt);
    XtFree((char *)pos_list);
  }
}

void toggle2_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  trans2_struct  *trans2;
  int            which;

  which=(int) client_data;
  XtVaGetValues(widget, XmNuserData, &trans2, NULL);

  if (XmToggleButtonGadgetGetState(widget)) 
    trans2->transform=which;
}

void trans2_ok_cb(Widget widget, XtPointer client_data, XtPointer call_data) {
  trans2_struct  *trans2;
  int            i, j, cols, rows;
  short          *col_width, is_numeric;
  float          var1, var2, result;
  String         **cells;
  char           *pcells, cell_cont[256];

  extern void showmess(Widget widget, char *title, char *text);

  XtVaGetValues(widget, XmNuserData, &trans2, NULL);
  XtVaGetValues(StatData, XmNcolumns, &cols, NULL);
  XtVaGetValues(StatData, XmNrows, &rows, NULL);

  if ((trans2->nlist1 == trans2->nlist2) &&
      (trans2->nrlist == trans2->nlist1) &&
      (trans2->nrlist == trans2->nlist2) ) {
    switch(trans2->transform) {
    case 0:   // +;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric) {
	    result=var1+var2;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      delete[]col_width;
      break;
    case 1:   // -;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric) {
	    result=var1-var2;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      delete[]col_width;
      break;
    case 2:   // *;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric) {
	    result=var1*var2;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      delete[]col_width;
      break;
    case 3:   // /;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric && var2!=0) {
	    result=var1/var2;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      
      delete[]col_width;
      break;
    case 4:   // 100*v2/v1;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric && var1!=0) {
	    result=100*var2/var1;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      delete[]col_width;
      break;
    case 5:   // 100*(v2-v1)/v1;
      col_width= new short[trans2->nrlist];
      for (i=0; i<trans2->nrlist; i++) col_width[i]=(short)COLWIDTH;
      XbaeMatrixAddColumns(StatData, cols, NULL, trans2->rlist, col_width,
      			   NULL, NULL, NULL, NULL, trans2->nrlist);
      XtVaGetValues(StatData, XmNcells, &cells, NULL);
      for (i=0; i<trans2->nrlist; i++) {
	for (j=0; j<rows; j++) {
	  is_numeric=1;
	  var1=strtod(cells[j][trans2->list1[i]], &pcells);
	  if (var1==0 && !strcmp(pcells,cells[j][trans2->list1[i]])) is_numeric=0;
	  var2=strtod(cells[j][trans2->list2[i]], &pcells);
	  if (var2==0 && !strcmp(pcells,cells[j][trans2->list2[i]])) is_numeric=0;
	  if (is_numeric && var1!=0) {
	    result=100*(var2-var1)/var1;
	    sprintf(cell_cont, "%f", result);
	  }
	  else {
	    strcpy(cell_cont, "ndef");
	  }
	  XbaeMatrixSetCell(StatData, j, cols+i, cell_cont);
	}
      }
      delete[]col_width;
      break;
    }
    XtUnmanageChild(widget);
  }
  else {
    showmess(widget, TRANS2_T, TRANS2_1_M);
  }
}
